/*The command to run the demo is as follows:

mpiexec -n 2 env LD_LIBRARY_PATH=/home/nbeers/visit060327/lib VISITPLUGINDIR=/home/nbeers/visit060327/plugins ./demo

The section of that command '-n 2' defines the number of processes the program starts. Changing the number changes the number of processes.
*/
#include "demo.H"

bool quitFlag;
int commandFlag;

void ProcessCommand(string cmd);
void ControlCommandCallback(const char *cmd, int int_data, float float_data, const char *string_data);
Attribute *squareData(Attribute *data, string &newName, string &meshName);

int main(int argc, char *argv[]){
  //Initialize Basin (Step 1)
    int numprocs,myid;
    basin_initialize(myid,numprocs);
    berr << "-------VisIt Demo for BASIN Data-------\n\n";

    MPI_Barrier(MPI_COMM_WORLD);

    //Reading files to BASIN Regions and copying data into grids and lists
    //(Step 2)
    Region cluster("cluster1.dat");
    List *l1 = cluster.get_list("RegionList");
    Region den("densfield.dat");
    Grid* g1 = den.get_grid();
    Geometry geom;

    vector<Attribute*> attrPtrs;

    //extracting Attributes from list
    Attribute *X= new Attribute(l1->get_attribute("x"));
    Attribute *Y= new Attribute(l1->get_attribute("y"));
    Attribute *Z= new Attribute(l1->get_attribute("z"));
    string meshname = "mesh", varname ="Mass";

    //exposing data for access to visit
    if(makePointMesh(meshname,X,Y,Z)!=0) //creates a Mesh
      berr <<"The mesh \""<<meshname<<"\" already exists. No mesh was created.\n";
    Attribute *data= new Attribute(l1->get_attribute("mass")); //creates a variable
    attrPtrs.push_back(data);
    if(makeVariable(varname,data,meshname)!=0)
      berr <<"The variable \""<<varname<<"\" already exists. No variable was created.\n";
    vector<int> grid(3,40);
    geom = *(cluster.get_geometry());
    vector<float> ll = geom.get_bound(0);
    vector<float> ur = geom.get_bound(1);
    varname = "mass_dens"; meshname = "griddedMesh";
    Attribute mass;
    Attribute *density =new Attribute(compute_density(*X,*Y,*Z,mass,grid,ll,ur,0));
    attrPtrs.push_back(density);
    //creating a grid using list data
    if(makeGridMesh(meshname,ll,ur,grid))
      berr <<"The mesh \""<<meshname<<"\" already exists. No mesh was created.\n";
    if( makeVariable(varname, density, meshname) )
      berr <<"The variable \""<<varname<<"\" already exists. No variable was created.\n";
    Attribute *d1=new Attribute(g1->get_attribute("density"));
    geom=*(den.get_geometry());
    ll = geom.get_bound(0);
    ur = geom.get_bound(1);
    vector<int> nx = g1->get_nx();
    varname = "density"; meshname = "grid";
    attrPtrs.push_back(d1);
    if(makeGridMesh(meshname,ll,ur,nx)) //creating grid from a grid file
      berr <<"The mesh \""<<meshname<<"\" already exists. No mesh was created.\n";
    if(makeVariable(varname,d1,meshname))//variable for above grid
      berr <<"The variable \""<<varname<<"\" already exists. No variable was created.\n";

    //set up the environment (Step 3)
    SetupEnvironment();

    {
      //Create and process the list of exposed functions
      //IMPORTANT NOTE: The maximum number of commands that can
      //      be displayed on the simulation window is 6. Also,
      //      exposing 6 names thru the command setGenCMD prevents
      //      the Custom Command button from showing up in the
      //      simulation window.
      vector<string> names(5);
      names[0]="checkLast";
      names[1]="2root";
      names[2]="square";
      names[3]="Multiply";
      names[4]="addPts";
      setGenCMD(names);
    }

    if(myid==0){
      VisItInitializeSocketAndDumpSimFile("Demo","Dynamic Data simulation","/no/useful/path",NULL,"SimGUI_1.ui");
      cerr << "Known commands:\n";
      cerr << "\t2root  :\tsquare roots newest variable\n";
      cerr << "\tsquare :\tsquares newest variable\n";
      cerr << "\tcheck  :\tStops a run\n";
      cerr << "\tadd    :\tAdd new data points to all variables\n";
      cerr << "\tquit   :\texit code\n";
    }
    berr << "command> ";
    //Loop that contains code for connecting to VisIt (Step 4)
    //IMPORTANT: The manner in which one breaks out of the loop can be triggered any way that you, the designer, see fit.
    //           This demo happens to make use of a global boolean flag "quitFlag" to do so. The manner in which the flag is triggered
    //           requires manual input from the user at the console. There are other methods for exiting this loop. (See below for 1 suggestion).
    while(!quitFlag){
        int visitstate;
	if(myid==0){
	  int blocking = (int) !runFlag;
	  visitstate = VisItDetectInput(blocking, fileno(stdin));
	}

	MPI_Bcast(&visitstate, 1, MPI_INT, 0, MPI_COMM_WORLD);
	commandFlag = -1;
	switch(visitstate){
	  case -5:
	  case -4:
	  case -3:
	  case -2:
	  case -1:
	    cerr << "\nUnrecoverable error... exiting.\n";
	    quitFlag = true;
	    break;
	  case 0:
	    break;
	  case 1:
	    //The VisIt connection actually occurs here.
	    //Without this case, nothing VisIt related will be able to occur.
	    //runFlag = false;
	    if(!VisItAttemptToCompleteConnection()){
	      string err = VisItGetLastError();
	      if(err.length() > 0)
		cerr << err << endl;
	      else
		cerr << "VisIt failed to connect successfully\n";
	    }else{
	        cerr << "Connected Successfully!\n";
		//These two functions expose commands that VisIt uses to
		//send information back. The name inside the parentheses
		//of the first function is declared in a different file.
		VisItSetSlaveProcessCallback(SlaveProcessCallback);
		VisItSetCommandCallback(ControlCommandCallback);
	    }
	    break;
	  case 2:
	    //Depending on what is going on, you may only need to connect once at this point and want your code to continue on once VisIt has disconnected.
	    //If that is the case, a command to alter the quitFlag's value to true here would do just that.
	    //berr << "Visit input detected. Processing VisIt Command. ";
	    if(!ProcessVisItCommand())
	      VisItDisconnect();
	    berr << endl;
	    break;
	  case 3:
	    {
	      char buff[100];
	      string input;
	      if(myid==0){
		cin >> input;
		if(cin.eof()){
		  input = "quit";
		  cerr << "quit\n";
		}
	      }
	      if(input.length() > 0 && input[input.length()-1] == '\n')
		input[input.length()-1] = '\0';
	      if(numprocs>1){
		strcpy(buff,input.c_str());
		MPI_Bcast(buff,100,MPI_CHAR,0,MPI_COMM_WORLD);
		if(myid!=0)
		  input = string(buff);
	      }
	      //cerr << myid << ": input = " << input << endl;
	      ProcessCommand(input);
	    }
	    break;
	}
	//The following if statement is used to make the proper function call.
	if(!quitFlag && (commandFlag>=0)){
	  switch(commandFlag){
	  case 0:
	    time_t starttime = time(NULL);
	    berr << "Pause..";
	    while(time(NULL)==starttime){
	      //pause;
	    }
	    berr << "..GO!\n";
	    break;
	  case 1:
	    //squareRootData();
	    StopSimulation();
	    berr <<"(NI)  Data square rooted.\n";
	    break;
	  case 2:
	    {
	      stringstream name;
	      name << "mass" << attrPtrs.size();
	      string temp1="mesh", temp2 = name.str();
	      attrPtrs.push_back(squareData(attrPtrs[0],temp2,temp1));
	      berr <<"  Data squared.\n";
	    }
	    StopSimulation();
	    break;
	  case 3:
	    //multiplyData();
	    {
	      int multiplier = 1;
	      berr <<"(NI)  Data multiplied by " << multiplier << ".\n";
	    }
	    StopSimulation();
	    break;
	  case 4:
	    berr <<"Making VisIt wait to plot.\n";
	    berr <<"Enter something and press enter to continue.  ";
	    if(myid==0){
	      string trash;
	      cin >> trash;
	    }
	    MPI_Barrier(MPI_COMM_WORLD);
	    berr <<"  Check finished.\n";
	    StopSimulation();
	    break;
	  case 5:
	    //runFlag = false;
	    berr <<"(NI)  Data added here.\n";//addPoints();
	    StopSimulation();
	    break;
	  case 6:
	    VisItDisconnect();
	    break;
	  case 7:
	    VisItAttemptToCompleteConnection();
	    break;
	  case 8:
	    VisItProcessEngineCommand();
	    break;
	  case 9:
	    
	    break;
	  default:
	    //cerr << cmd << " Is not a recognized command.\n";
	    berr << "  Custom command received. Processing...\n";
	    //if(CustomCommandParser(cmd))
	    //berr << cmd << "  is not a recognized command.\n";
	    berr << " Custom command processing not yet implemented.\n";
	  }
	}
	if(!quitFlag && !runFlag && visitstate==3){
	  berr << "command> ";
	}

    }
    cerr << "Good-bye\n";
    //Clean up BASIN and end main(). (Step 5)
    basin_finalize();
    return 0;
}


//Purpose: interprets commands input by the console or sent by VisIt.
//Each possible console input is assigned a number. When the commands are received from either the console or VisIt, they are sent thru ProcessCommand. That alters the variable commandFlag. Altering that variable allows the simulation above to access the function which corresponds to the selected commandFlag value. It also reduces the number of global variables required by the program.
void ProcessCommand(string cmd){
  //cerr << myid << " started processing the command.\n";
    if(cmd.compare("quit")==0)
      quitFlag = true;
    else if(cmd.compare("halt")==0)
      commandFlag = 0;
    else if(cmd.compare("2root")==0)
      commandFlag = 1;
    else if(cmd.compare("square")==0)
      commandFlag = 2;
    else if(cmd.compare("mult")==0 || cmd.compare("Multiply")==0)
      commandFlag = 3;
    else if(cmd.compare("check")==0 || cmd.compare("checkLast")==0)
      commandFlag = 4;
    else if(cmd.compare("add")==0 || cmd.compare("addPts")==0)
      commandFlag = 5;
    else if(cmd.compare("visit_disconnect")==0)
      commandFlag = 6;
    else if(cmd.compare("visit_connect")==0)
      commandFlag = 7;
    else if(cmd.compare("visit_process")==0)
      commandFlag = 8;
    else if(cmd.compare("")==0)
      commandFlag = 9;
    else
      commandFlag = 100;
}

//Create ControlCommandCallback(). (Step 6)
void ControlCommandCallback(const char *cmd, int int_data, float float_data, const char *string_data){
    berr << " ControlCommandCallback(cmd=" << cmd << ")\n";
    string buff = string(cmd);
    ProcessCommand(buff);
}

Attribute *squareData(Attribute *data, string &newName, string &meshName){
    Attribute *var = new Attribute(pow(*data + 0, 2.0));
    makeVariable(newName, data, meshName);
    return var;
}
